import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;

import com.aliyun.odps.data.TableInfo;
import com.aliyun.odps.graph.Aggregator;
import com.aliyun.odps.graph.ComputeContext;
import com.aliyun.odps.graph.GraphJob;
import com.aliyun.odps.graph.GraphLoader;
import com.aliyun.odps.graph.MutationContext;
import com.aliyun.odps.graph.Vertex;
import com.aliyun.odps.graph.WorkerContext;
import com.aliyun.odps.io.Writable;
import com.aliyun.odps.io.NullWritable;
import com.aliyun.odps.io.LongWritable;
import com.aliyun.odps.io.DoubleWritable;
import com.aliyun.odps.io.WritableRecord;

import no.uib.cipr.matrix.*;

public class KmeansClustring {

  public static class KmeansValue implements Writable {
    
    DenseVector sample;
    
    public KmeansValue() { 
    }
    
    public KmeansValue(DenseVector v) {
      this.sample = v;
    }

    @Override
    public void write(DataOutput out) throws IOException {
      wirteForDenseVector(out, sample);
      
    }

    @Override
    public void readFields(DataInput in) throws IOException {
      sample = readFieldsForDenseVector(in);
    }
  }
  
  public static class KmeansReader extends 
    GraphLoader<LongWritable, KmeansValue, NullWritable, NullWritable> {
    
    @Override
    public void load(
        LongWritable recordNum,
        WritableRecord record,
        MutationContext<LongWritable, KmeansValue, NullWritable, NullWritable> context)
        throws IOException {
      KmeansVertex v = new KmeansVertex();
      v.setId(recordNum);
      
      int n = record.size();
      DenseVector dv = new DenseVector(n);
      for (int i = 0; i < n; i++) {
        dv.set(i, ((DoubleWritable)record.get(i)).get());
      }
      v.setValue(new KmeansValue(dv));
      
      context.addVertexRequest(v);
    }
  }
  
  public static class KmeansVertex extends
    Vertex<LongWritable, KmeansValue, NullWritable, NullWritable> {

    @Override
    public void compute(
        ComputeContext<LongWritable, KmeansValue, NullWritable, NullWritable> context,
        Iterable<NullWritable> messages) throws IOException {
      context.aggregate(getValue());
    }
  }
  
  public static class KmeansAggrValue implements Writable {
    
    DenseMatrix centroids;
    DenseMatrix sums; // used to recalculate new centroids
    DenseVector counts; // used to recalculate new centroids

    @Override
    public void write(DataOutput out) throws IOException {
      wirteForDenseDenseMatrix(out, centroids);
      wirteForDenseDenseMatrix(out, sums);
      wirteForDenseVector(out, counts);
    }

    @Override
    public void readFields(DataInput in) throws IOException {
      centroids = readFieldsForDenseMatrix(in);
      sums = readFieldsForDenseMatrix(in);
      counts = readFieldsForDenseVector(in);
    }
  }
  
  @SuppressWarnings("rawtypes")
  public static class KmeansAggregator extends Aggregator<KmeansAggrValue> {

    public KmeansAggrValue createStartupValue(WorkerContext context) throws IOException {
      KmeansAggrValue av = new KmeansAggrValue();
      
      byte[] centers = context.readCacheFile("centers");
      String lines[] = new String(centers).split("\n");
      
      int rows = lines.length;
      int cols = lines[0].split(",").length; // assumption rows >= 1 
      
      av.centroids = new DenseMatrix(rows, cols);
      av.sums = new DenseMatrix(rows, cols);
      av.sums.zero();
      av.counts = new DenseVector(rows);
      av.counts.zero();
      
      for (int i = 0; i < lines.length; i++) {
        String[] ss = lines[i].split(",");
        for (int j = 0; j < ss.length; j++) {
          av.centroids.set(i, j, Double.valueOf(ss[j]));
        }
      }
      return av;
    }
    
    @Override
    public KmeansAggrValue createInitialValue(WorkerContext context)
        throws IOException {
      KmeansAggrValue av = (KmeansAggrValue)context.getLastAggregatedValue(0);
      
      // reset for next iteration
      av.sums.zero();
      av.counts.zero();
      
      return av;
    }

    @Override
    public void aggregate(KmeansAggrValue value, Object item)
        throws IOException {
      DenseVector sample = ((KmeansValue)item).sample;
      
      // find the nearest centroid
      int min = findNearestCentroid(value.centroids, sample);
      
      // update sum and count
      for (int i = 0; i < sample.size(); i ++) {
        value.sums.add(min, i, sample.get(i));
      }
      value.counts.add(min, 1.0d);
    }
    
    static int findNearestCentroid(DenseMatrix centroids, DenseVector point) {
      int min = 0;
      double minDist = Double.MAX_VALUE;
      for (int i = 0; i < centroids.numRows(); i++) {
        // squared Euclidean distance
        double dist = 0.0d;
        for (int j = 0; j < centroids.numColumns(); j++) {
          double d = point.get(j) - centroids.get(i, j);
          dist += d * d;
        }
        
        if (dist <= minDist) {
          minDist = dist;
          min = i;
        }
      }
      
      return min;
    }

    @Override
    public void merge(KmeansAggrValue value, KmeansAggrValue partial)
        throws IOException {
      value.sums.add(partial.sums);
      value.counts.add(partial.counts);
    }

    @Override
    public boolean terminate(WorkerContext context, KmeansAggrValue value)
        throws IOException {
      // Calculate the new means to be the centroids (original sums)
      DenseMatrix newCentriods = calculateNewCentroids(value.sums, value.counts, value.centroids);
      
      // print old centroids and new centroids for debugging
      System.out.println("\nsuperstep: " + context.getSuperstep() + 
          "\nold centriod:\n" + value.centroids + " new centriod:\n" + newCentriods);
      
      boolean converged = isConverged(newCentriods, value.centroids, 0.05d);
      System.out.println("superstep: " + context.getSuperstep() + "/" 
          + (context.getMaxIteration() - 1) + " converged: " + converged);
      if (converged || context.getSuperstep() == context.getMaxIteration() - 1) {
        // converged or reach max iteration, output centriods
        for (int i = 0; i < newCentriods.numRows(); i++) {
          Writable[] centriod = new Writable[newCentriods.numColumns()];
          for (int j = 0; j < newCentriods.numColumns(); j++) {
            centriod[j] = new DoubleWritable(newCentriods.get(i, j));
          }
          context.write(centriod);
        }
        
        // true means to terminate iteration
        return true;
      }
      
      // update centriods
      value.centroids.set(newCentriods);
      // false means to continue iteration
      return false;
    }
    
    // sums = sums/counts
    static DenseMatrix calculateNewCentroids(DenseMatrix sums, DenseVector counts, DenseMatrix old) {
      for (int i = 0; i < sums.numRows(); i ++) {
        if (Math.abs(counts.get(i)) > 0.5d) { // count is zero, mean=sum/count
          for (int j = 0; j < sums.numColumns(); j++) {
            double mean = sums.get(i, j) / counts.get(i);
            sums.set(i, j, mean);
          }
        } else {  // count is above zero, new centriod equals old one
          for (int j = 0; j < sums.numColumns(); j++) {
            sums.set(i, j, old.get(i, j));
          }
        }
      }
      return sums;
    }
    
    static boolean isConverged(DenseMatrix newC, DenseMatrix oldC, double threshold) {
      for (int i = 0; i < newC.numRows(); i++) {
        double dist = 0.0d;
        for (int j = 0; j < newC.numColumns(); j++) {
          double d = newC.get(i, j) - oldC.get(i, j);
          dist += d * d;
        }
        if (Math.sqrt(dist) > threshold) {
          return false;
        }
      }
      return true;
    }

  }
  
  public static void wirteForDenseVector(DataOutput out, DenseVector v) throws IOException {
    int n = v.size();
    out.writeInt(n);
    for (int i = 0; i < n; i++) {
      out.writeDouble(v.get(i));
    }
  }
  
  public static DenseVector readFieldsForDenseVector(DataInput in) throws IOException {
    int n = in.readInt();
    
    DenseVector v = new DenseVector(n);
    for (int i = 0; i < n; i++) {
      v.set(i, in.readDouble());
    }
    
    return v;
  }
  
  public static void wirteForDenseDenseMatrix(DataOutput out, DenseMatrix m) throws IOException {
    int rows = m.numRows();
    int cols = m.numColumns();
    
    out.writeInt(rows);
    out.writeInt(cols);
    for (int i = 0; i < rows; i++) {
      for (int j = 0; j < cols; j++) {
        out.writeDouble(m.get(i, j));
      }
    }
  }
  
  public static DenseMatrix readFieldsForDenseMatrix(DataInput in) throws IOException {
    int rows = in.readInt();
    int cols = in.readInt();
    DenseMatrix m = new DenseMatrix(rows, cols);
    
    for (int i = 0; i < rows; i++) {
      for (int j = 0; j < cols; j++) {
        m.set(i, j, in.readDouble());
      }
    }
    
    return m;
  }
  

  private static void printUsage() {
    System.out.println("Usage: <in> <out> [Max iterations (default 30)]");
    System.exit(-1);
  }

  public static void main(String[] args) throws IOException {
    if (args.length < 2)
      printUsage();

    GraphJob job = new GraphJob();

    job.setGraphLoaderClass(KmeansReader.class);
    job.setRuntimePartitioning(false);
    job.setVertexClass(KmeansVertex.class);
    job.setAggregatorClass(KmeansAggregator.class);
    job.addInput(TableInfo.builder().tableName(args[0]).build());
    job.addOutput(TableInfo.builder().tableName(args[1]).build());

    // default max iteration is 30
    job.setMaxIteration(30);
    if (args.length >= 3)
      job.setMaxIteration(Integer.parseInt(args[2]));

    long start = System.currentTimeMillis();
    job.run();
    System.out.println("Job Finished in "
        + (System.currentTimeMillis() - start) / 1000.0 + " seconds");
  }
  
}
